# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from inspect import isawaitable

from newrelic.api.function_trace import FunctionTrace, function_trace
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import current_transaction
from newrelic.api.web_transaction import web_transaction
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import function_wrapper, wrap_function_wrapper


def _bind_add(uri, methods, handler, *args, **kwargs):
    return uri, methods, handler, args, kwargs


@function_wrapper
def _nr_wrapper_handler_(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)
    view_class = getattr(wrapped, "view_class", None)
    view = view_class or wrapped
    if view_class:
        try:
            method = args[0].method.lower()
            name = callable_name(view_class) + "." + method
            view = getattr(view_class, method)
        except:
            pass

    transaction.set_transaction_name(name, priority=3)
    import sanic

    transaction.add_framework_info(name="Sanic", version=sanic.__version__)

    with FunctionTrace(name=name, source=view):
        return wrapped(*args, **kwargs)


@function_wrapper
def _nr_sanic_router_add(wrapped, instance, args, kwargs):
    uri, methods, handler, args, kwargs = _bind_add(*args, **kwargs)

    # Cache the callable_name on the handler object
    callable_name(handler)
    if hasattr(wrapped, "view_class"):
        callable_name(wrapped.view_class)
    wrapped_handler = _nr_wrapper_handler_(handler)

    return wrapped(uri, methods, wrapped_handler, *args, **kwargs)


@function_wrapper
def _nr_sanic_router_get(wrapped, instance, args, kwargs):
    # Rename all transactions that generate an exception in the router get
    try:
        return wrapped(*args, **kwargs)
    except Exception:
        transaction = current_transaction()
        if transaction:
            name = callable_name(wrapped)
            transaction.set_transaction_name(name, priority=2)
        raise


@function_wrapper
def _nr_wrapper_error_handler_(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if not transaction:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)
    transaction.set_transaction_name(name, priority=1)
    try:
        response = function_trace(name=name)(wrapped)(*args, **kwargs)
    except:
        notice_error()
        raise

    return response


def _bind_error_add(exception, handler, *args, **kwargs):
    return exception, handler


@function_wrapper
def _nr_sanic_error_handlers(wrapped, instance, args, kwargs):
    exception, handler = _bind_error_add(*args, **kwargs)

    # Cache the callable_name on the handler object
    callable_name(handler)
    wrapped_handler = _nr_wrapper_error_handler_(handler)

    return wrapped(exception, wrapped_handler)


@function_wrapper
def error_response(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    exc_info = sys.exc_info()
    try:
        response = wrapped(*args, **kwargs)
    except:
        notice_error(exc_info)
        # We record the original exception and the exception generated by the
        # error handler
        notice_error()
        raise
    else:
        # response can be a response object or a coroutine
        if hasattr(response, "status"):
            notice_error(error=exc_info, status_code=response.status)
        else:
            notice_error(exc_info)
    finally:
        exc_info = None

    return response


def _sanic_app_init(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)

    error_handler = getattr(instance, "error_handler")
    if hasattr(error_handler, "response"):
        instance.error_handler.response = error_response(error_handler.response)
    if hasattr(error_handler, "add"):
        error_handler.add = _nr_sanic_error_handlers(error_handler.add)

    router = getattr(instance, "router")
    if hasattr(router, "add"):
        router.add = _nr_sanic_router_add(router.add)
    if hasattr(router, "get"):
        # Cache the callable_name on the router.get
        callable_name(router.get)
        router.get = _nr_sanic_router_get(router.get)

    return result


def _nr_sanic_response_get_headers(wrapped, instance, args, kwargs):
    # We call the wrapped function twice to allow the function to mutate the headers
    result = wrapped(*args, **kwargs)
    transaction = current_transaction()

    if transaction is None:
        return result

    # instance is the response object
    cat_headers = transaction.process_response(str(instance.status), instance.headers.items())

    for header_name, header_value in cat_headers:
        if header_name not in instance.headers:
            instance.headers[header_name] = header_value

    return wrapped(*args, **kwargs)


async def _nr_sanic_response_send(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    result = wrapped(*args, **kwargs)
    if isawaitable(result):
        await result

    if transaction is None:
        return result

    # instance is the response object
    cat_headers = transaction.process_response(str(instance.status), instance.headers.items())

    for header_name, header_value in cat_headers:
        if header_name not in instance.headers:
            instance.headers[header_name] = header_value

    return result


def _nr_sanic_response_parse_headers(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    # instance is the response object
    cat_headers = transaction.process_response(str(instance.status), instance.headers.items())

    for header_name, header_value in cat_headers:
        if header_name not in instance.headers:
            instance.headers[header_name] = header_value

    return wrapped(*args, **kwargs)


def _nr_wrapper_middleware_(attach_to):
    is_request_middleware = attach_to == "request"

    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        name = callable_name(wrapped)
        if is_request_middleware:
            transaction.set_transaction_name(name, priority=2)
        response = function_trace(name=name)(wrapped)(*args, **kwargs)

        return response

    return _wrapper


def _bind_middleware(middleware, attach_to="request", *args, **kwargs):
    return middleware, attach_to


def _nr_sanic_register_middleware_(wrapped, instance, args, kwargs):
    middleware, attach_to = _bind_middleware(*args, **kwargs)

    # Cache the callable_name on the middleware object
    callable_name(middleware)
    middleware_func = middleware
    if hasattr(middleware, "func"):
        name = callable_name(middleware.func)
        middleware_func = middleware.func

    wrapped_middleware = _nr_wrapper_middleware_(attach_to)(middleware_func)
    wrapped(wrapped_middleware, attach_to)
    return middleware


def _bind_request(request, *args, **kwargs):
    return request


def _nr_sanic_transaction_wrapper_(wrapped, instance, args, kwargs):
    request = _bind_request(*args, **kwargs)
    # If the request is a websocket request do not wrap it
    if request.headers.get("upgrade", "").lower() == "websocket":
        return wrapped(*args, **kwargs)

    return web_transaction(
        request_method=request.method,
        request_path=request.path,
        query_string=request.query_string,
        headers=request.headers,
    )(wrapped)(*args, **kwargs)


def _nr_wrap_touchup_run(wrapped, instance, args, kwargs):
    # TouchUp uses metaprogramming to rewrite methods of classes on startup.
    # To properly wrap them we need to catch the call to TouchUp.run and
    # reinstrument any methods that were replaced with uninstrumented versions.

    result = wrapped(*args, **kwargs)

    if "sanic.app" in sys.modules:
        module = sys.modules["sanic.app"]
        target = args[0]

        if isinstance(target, module.Sanic):
            # Reinstrument class after metaclass "TouchUp" has finished rewriting methods on the class.
            target_cls = module.Sanic
            if hasattr(target_cls, "handle_request") and not hasattr(target_cls.handle_request, "__wrapped__"):
                wrap_function_wrapper(module, "Sanic.handle_request", _nr_sanic_transaction_wrapper_)

    return result


def instrument_sanic_app(module):
    wrap_function_wrapper(module, "Sanic.handle_request", _nr_sanic_transaction_wrapper_)
    wrap_function_wrapper(module, "Sanic.__init__", _sanic_app_init)
    wrap_function_wrapper(module, "Sanic.register_middleware", _nr_sanic_register_middleware_)
    if hasattr(module.Sanic, "register_named_middleware"):
        wrap_function_wrapper(module, "Sanic.register_named_middleware", _nr_sanic_register_middleware_)


def instrument_sanic_response(module):
    if hasattr(module.BaseHTTPResponse, "send"):
        wrap_function_wrapper(module, "BaseHTTPResponse.send", _nr_sanic_response_send)
    else:
        if hasattr(module.BaseHTTPResponse, "get_headers"):
            wrap_function_wrapper(module, "BaseHTTPResponse.get_headers", _nr_sanic_response_get_headers)
        if hasattr(module.BaseHTTPResponse, "_parse_headers"):
            wrap_function_wrapper(module, "BaseHTTPResponse._parse_headers", _nr_sanic_response_parse_headers)


def instrument_sanic_touchup_service(module):
    if hasattr(module, "TouchUp") and hasattr(module.TouchUp, "run"):
        wrap_function_wrapper(module, "TouchUp.run", _nr_wrap_touchup_run)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from newrelic.common.encoding_utils import deobfuscate, json_decode
from newrelic.common.object_wrapper import transient_function_wrapper


def validate_browser_attributes(required_params=None, forgone_params=None):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    @transient_function_wrapper("newrelic.api.web_transaction", "WSGIWebTransaction.browser_timing_footer")
    def _validate_browser_attributes(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise

        # pick out attributes from footer string_types

        footer_data = result.split("NREUM.info=")[1]
        footer_data = footer_data.split("</script>")[0]
        footer_data = json.loads(footer_data)

        if "intrinsic" in required_params:
            for attr in required_params["intrinsic"]:
                assert attr in footer_data

        if "atts" in footer_data:
            obfuscation_key = instance._settings.license_key[:13]
            attributes = json_decode(deobfuscate(footer_data["atts"], obfuscation_key))
        else:

            # if there are no user or agent attributes, there will be no dict
            # for them in the browser data

            attributes = None

        if "user" in required_params:
            for attr in required_params["user"]:
                assert attr in attributes["u"]

        if "agent" in required_params:
            for attr in required_params["agent"]:
                assert attr in attributes["a"]

        if "user" in forgone_params:
            if attributes:
                if "u" in attributes:
                    for attr in forgone_params["user"]:
                        assert attr not in attributes["u"]

        if "agent" in forgone_params:
            if attributes:
                if "a" in attributes:
                    for attr in forgone_params["agent"]:
                        assert attr not in attributes["a"]

        return result

    return _validate_browser_attributes



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' The purpose of these tests is to confirm that we will record
record instance info for Redis Blaster commands that go through
redis.Connection:send_command(). Commands that don't use send_command,
like the one that use the fanout client, won't have instance info.
'''

import pytest
import redis
import six

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import redis_settings
from testing_support.util import instance_hostname

DB_SETTINGS = redis_settings()[0]
REDIS_PY_VERSION = get_package_version_tuple("redis")


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

# We don't record instance metrics when using redis blaster,
# so we just check for base metrics.

_base_scoped_metrics = (
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
)

_base_rollup_metrics = (
        ('Datastore/all', 2),
        ('Datastore/allOther', 2),
        ('Datastore/Redis/all', 2),
        ('Datastore/Redis/allOther', 2),
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
)

_disable_rollup_metrics = list(_base_rollup_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Redis/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 2)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)


# Operations
def exercise_redis(routing_client):
    routing_client.set('key', 'value')
    routing_client.get('key')


def exercise_fanout(cluster):
    with cluster.fanout(hosts='all') as client:
        client.execute_command('CLIENT', 'LIST')


# Tests
@pytest.mark.skipif(six.PY3, reason='Redis Blaster is Python 2 only.')
@pytest.mark.skipif(REDIS_PY_VERSION < (2, 10, 2),
        reason='Redis Blaster requires redis>=2.10.2')
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_rb:test_redis_blaster_operation_enable_instance',
        scoped_metrics=_base_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@background_task()
def test_redis_blaster_operation_enable_instance():
    from rb import Cluster

    cluster = Cluster(
            hosts={0: {'port': DB_SETTINGS['port']}},
            host_defaults={'host': DB_SETTINGS['host']}
    )
    exercise_fanout(cluster)

    client = cluster.get_routing_client()
    exercise_redis(client)


@pytest.mark.skipif(six.PY3, reason='Redis Blaster is Python 2 only.')
@pytest.mark.skipif(REDIS_PY_VERSION < (2, 10,2 ),
        reason='Redis Blaster requires redis>=2.10.2')
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_rb:test_redis_blaster_operation_disable_instance',
        scoped_metrics=_base_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@background_task()
def test_redis_blaster_operation_disable_instance():
    from rb import Cluster

    cluster = Cluster(
            hosts={0: {'port': DB_SETTINGS['port']}},
            host_defaults={'host': DB_SETTINGS['host']}
    )
    exercise_fanout(cluster)

    client = cluster.get_routing_client()
    exercise_redis(client)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
from testing_support.fixtures import override_generic_settings

from newrelic.core.config import global_settings
from newrelic.core.environment import environment_settings

settings = global_settings()


def module(version):
    class Module(object):
        pass

    if version:
        Module.__version__ = version

    return Module


def test_plugin_list():
    # Let's pretend we fired an import hook
    import newrelic.hooks.adapter_gunicorn  # noqa: F401

    environment_info = environment_settings()

    for key, plugin_list in environment_info:
        if key == "Plugin List":
            break
    else:
        assert False, "'Plugin List' not found"

    # Check that bogus plugins don't get reported
    assert "newrelic.hooks.newrelic" not in plugin_list
    # Check that plugin that should get reported has version info.
    assert "pytest (%s)" % (pytest.__version__) in plugin_list


@override_generic_settings(settings, {"package_reporting.enabled": False})
def test_plugin_list_when_package_reporting_disabled():
    # Let's pretend we fired an import hook
    import newrelic.hooks.adapter_gunicorn  # noqa: F401

    environment_info = environment_settings()

    for key, plugin_list in environment_info:
        if key == "Plugin List":
            break
    else:
        assert False, "'Plugin List' not found"

    # Check that bogus plugins don't get reported
    assert plugin_list == []


class NoIteratorDict(object):
    def __init__(self, d):
        self.d = d

    def copy(self):
        return self.d.copy()

    def get(self, *args, **kwargs):
        return self.d.get(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self.d.__getitem__(*args, **kwargs)

    def __contains__(self, *args, **kwargs):
        return self.d.__contains__(*args, **kwargs)


def test_plugin_list_uses_no_sys_modules_iterator(monkeypatch):
    modules = NoIteratorDict(sys.modules)
    monkeypatch.setattr(sys, "modules", modules)

    # If environment_settings iterates over sys.modules, an attribute error will be generated
    environment_info = environment_settings()

    for key, plugin_list in environment_info:
        if key == "Plugin List":
            break
    else:
        assert False, "'Plugin List' not found"


@pytest.mark.parametrize(
    "loaded_modules,dispatcher,dispatcher_version,worker_version",
    (
        ({"uvicorn": module("4.5.6")}, "uvicorn", "4.5.6", None),
        (
            {
                "gunicorn": module("1.2.3"),
                "uvicorn": module("4.5.6"),
                "uvicorn.workers": object(),
            },
            "gunicorn (uvicorn)",
            "1.2.3",
            "4.5.6",
        ),
        ({"uvicorn": object()}, "uvicorn", None, None),
        (
            {
                "gunicorn": object(),
                "uvicorn": module("4.5.6"),
                "uvicorn.workers": object(),
            },
            "gunicorn (uvicorn)",
            None,
            "4.5.6",
        ),
        (
            {"gunicorn": module("1.2.3"), "uvicorn": None, "uvicorn.workers": object()},
            "gunicorn (uvicorn)",
            "1.2.3",
            None,
        ),
        (
            {"gunicorn": object(), "uvicorn": object(), "uvicorn.workers": object()},
            "gunicorn (uvicorn)",
            None,
            None,
        ),
    ),
)
def test_uvicorn_dispatcher(monkeypatch, loaded_modules, dispatcher, dispatcher_version, worker_version):
    # Let's pretend we load some modules
    for name, module in loaded_modules.items():
        monkeypatch.setitem(sys.modules, name, module)

    environment_info = environment_settings()

    actual_dispatcher = None
    actual_dispatcher_version = None
    actual_worker_version = None
    for key, value in environment_info:
        if key == "Dispatcher":
            assert actual_dispatcher is None
            actual_dispatcher = value
        elif key == "Dispatcher Version":
            assert actual_dispatcher_version is None
            actual_dispatcher_version = value
        elif key == "Worker Version":
            assert actual_worker_version is None
            actual_worker_version = value

    assert actual_dispatcher == dispatcher
    assert actual_dispatcher_version == dispatcher_version
    assert actual_worker_version == worker_version



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x08v1.proto\x12\x15\x63om.newrelic.trace.v1"7\n\tSpanBatch\x12*\n\x05spans\x18\x01 \x03(\x0b\x32\x1b.com.newrelic.trace.v1.Span"\x86\x04\n\x04Span\x12\x10\n\x08trace_id\x18\x01 \x01(\t\x12?\n\nintrinsics\x18\x02 \x03(\x0b\x32+.com.newrelic.trace.v1.Span.IntrinsicsEntry\x12H\n\x0fuser_attributes\x18\x03 \x03(\x0b\x32/.com.newrelic.trace.v1.Span.UserAttributesEntry\x12J\n\x10\x61gent_attributes\x18\x04 \x03(\x0b\x32\x30.com.newrelic.trace.v1.Span.AgentAttributesEntry\x1aX\n\x0fIntrinsicsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.com.newrelic.trace.v1.AttributeValue:\x02\x38\x01\x1a\\\n\x13UserAttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.com.newrelic.trace.v1.AttributeValue:\x02\x38\x01\x1a]\n\x14\x41gentAttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.com.newrelic.trace.v1.AttributeValue:\x02\x38\x01"t\n\x0e\x41ttributeValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x02 \x01(\x08H\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x42\x07\n\x05value"%\n\x0cRecordStatus\x12\x15\n\rmessages_seen\x18\x01 \x01(\x04\x32\xc5\x01\n\rIngestService\x12T\n\nRecordSpan\x12\x1b.com.newrelic.trace.v1.Span\x1a#.com.newrelic.trace.v1.RecordStatus"\x00(\x01\x30\x01\x12^\n\x0fRecordSpanBatch\x12 .com.newrelic.trace.v1.SpanBatch\x1a#.com.newrelic.trace.v1.RecordStatus"\x00(\x01\x30\x01\x62\x06proto3'
)


_SPANBATCH = DESCRIPTOR.message_types_by_name["SpanBatch"]
_SPAN = DESCRIPTOR.message_types_by_name["Span"]
_SPAN_INTRINSICSENTRY = _SPAN.nested_types_by_name["IntrinsicsEntry"]
_SPAN_USERATTRIBUTESENTRY = _SPAN.nested_types_by_name["UserAttributesEntry"]
_SPAN_AGENTATTRIBUTESENTRY = _SPAN.nested_types_by_name["AgentAttributesEntry"]
_ATTRIBUTEVALUE = DESCRIPTOR.message_types_by_name["AttributeValue"]
_RECORDSTATUS = DESCRIPTOR.message_types_by_name["RecordStatus"]
SpanBatch = _reflection.GeneratedProtocolMessageType(
    "SpanBatch",
    (_message.Message,),
    {
        "DESCRIPTOR": _SPANBATCH,
        "__module__": "v1_pb2"
        # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.SpanBatch)
    },
)
_sym_db.RegisterMessage(SpanBatch)

Span = _reflection.GeneratedProtocolMessageType(
    "Span",
    (_message.Message,),
    {
        "IntrinsicsEntry": _reflection.GeneratedProtocolMessageType(
            "IntrinsicsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _SPAN_INTRINSICSENTRY,
                "__module__": "v1_pb2"
                # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.Span.IntrinsicsEntry)
            },
        ),
        "UserAttributesEntry": _reflection.GeneratedProtocolMessageType(
            "UserAttributesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _SPAN_USERATTRIBUTESENTRY,
                "__module__": "v1_pb2"
                # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.Span.UserAttributesEntry)
            },
        ),
        "AgentAttributesEntry": _reflection.GeneratedProtocolMessageType(
            "AgentAttributesEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _SPAN_AGENTATTRIBUTESENTRY,
                "__module__": "v1_pb2"
                # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.Span.AgentAttributesEntry)
            },
        ),
        "DESCRIPTOR": _SPAN,
        "__module__": "v1_pb2"
        # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.Span)
    },
)
_sym_db.RegisterMessage(Span)
_sym_db.RegisterMessage(Span.IntrinsicsEntry)
_sym_db.RegisterMessage(Span.UserAttributesEntry)
_sym_db.RegisterMessage(Span.AgentAttributesEntry)

AttributeValue = _reflection.GeneratedProtocolMessageType(
    "AttributeValue",
    (_message.Message,),
    {
        "DESCRIPTOR": _ATTRIBUTEVALUE,
        "__module__": "v1_pb2"
        # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.AttributeValue)
    },
)
_sym_db.RegisterMessage(AttributeValue)

RecordStatus = _reflection.GeneratedProtocolMessageType(
    "RecordStatus",
    (_message.Message,),
    {
        "DESCRIPTOR": _RECORDSTATUS,
        "__module__": "v1_pb2"
        # @@protoc_insertion_point(class_scope:com.newrelic.trace.v1.RecordStatus)
    },
)
_sym_db.RegisterMessage(RecordStatus)

_INGESTSERVICE = DESCRIPTOR.services_by_name["IngestService"]
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    _SPAN_INTRINSICSENTRY._options = None
    _SPAN_INTRINSICSENTRY._serialized_options = b"8\001"
    _SPAN_USERATTRIBUTESENTRY._options = None
    _SPAN_USERATTRIBUTESENTRY._serialized_options = b"8\001"
    _SPAN_AGENTATTRIBUTESENTRY._options = None
    _SPAN_AGENTATTRIBUTESENTRY._serialized_options = b"8\001"
    _SPANBATCH._serialized_start = 35
    _SPANBATCH._serialized_end = 90
    _SPAN._serialized_start = 93
    _SPAN._serialized_end = 611
    _SPAN_INTRINSICSENTRY._serialized_start = 334
    _SPAN_INTRINSICSENTRY._serialized_end = 422
    _SPAN_USERATTRIBUTESENTRY._serialized_start = 424
    _SPAN_USERATTRIBUTESENTRY._serialized_end = 516
    _SPAN_AGENTATTRIBUTESENTRY._serialized_start = 518
    _SPAN_AGENTATTRIBUTESENTRY._serialized_end = 611
    _ATTRIBUTEVALUE._serialized_start = 613
    _ATTRIBUTEVALUE._serialized_end = 729
    _RECORDSTATUS._serialized_start = 731
    _RECORDSTATUS._serialized_end = 768
    _INGESTSERVICE._serialized_start = 771
    _INGESTSERVICE._serialized_end = 968
# @@protoc_insertion_point(module_scope)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from testing_support.fixtures import reset_core_stats_engine


@reset_core_stats_engine()
@background_task()
def test_callsite_parameter_processor(callsite_parameter_logger, structlog_caplog):
    callsite_parameter_logger.msg("Dog")
    assert "Dog" in structlog_caplog.caplog[0]
    assert "filename='test_structlog_processors.py'" in structlog_caplog.caplog[0]
    assert "func_name='test_callsite_parameter_processor'" in structlog_caplog.caplog[0]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import memcache
from newrelic.hooks.datastore_memcache import (_memcache_client_methods,
        _memcache_multi_methods)

def test_all_methods_wrapped():
    client = memcache.Client([])

    all_methods = (list(_memcache_client_methods) +
            list(_memcache_multi_methods))

    for m in all_methods:
        method = getattr(client, m)
        assert hasattr(method, '__wrapped__')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from io import BytesIO as IO
except ImportError:
    import StringIO as IO

import webtest

from flask import Flask
from flask import Response
from flask import send_file
from flask_compress import Compress

from newrelic.api.transaction import (get_browser_timing_header,
        get_browser_timing_footer)

application = Flask(__name__)

compress = Compress()
compress.init_app(application)


@application.route('/compress')
def index_page():
    return '<body>' + 500 * 'X' + '</body>'


@application.route('/html_insertion')
def html_insertion():
    return ('<!DOCTYPE html><html><head>Some header</head>'
           '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
           '</body></html>')


@application.route('/html_insertion_manual')
def html_insertion_manual():
    header = get_browser_timing_header()
    footer = get_browser_timing_footer()

    header = get_browser_timing_header()
    footer = get_browser_timing_footer()

    assert header == ''
    assert footer == ''

    return ('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')


@application.route('/html_insertion_unnamed_attachment_header')
def html_insertion_unnamed_attachment_header():
    response = Response(
            response='<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')
    response.headers.add('Content-Disposition',
                'attachment')
    return response


@application.route('/html_insertion_named_attachment_header')
def html_insertion_named_attachment_header():
    response = Response(
            response='<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')
    response.headers.add('Content-Disposition',
                'attachment; filename="X"')
    return response


@application.route('/html_served_from_file')
def html_served_from_file():
    file = IO()
    contents = b"""
    <!DOCTYPE html><html><head>Some header</head>
    <body><h1>My First Heading</h1><p>My first paragraph.</p>
    </body></html>
    """
    file.write(contents)
    file.seek(0)
    return send_file(file, mimetype='text/html')


@application.route('/text_served_from_file')
def text_served_from_file():
    file = IO()
    contents = b"""
    <!DOCTYPE html><html><head>Some header</head>
    <body><h1>My First Heading</h1><p>My first paragraph.</p>
    </body></html>
    """
    file.write(contents)
    file.seek(0)
    return send_file(file, mimetype='text/plain')


_test_application = webtest.TestApp(application)


@application.route('/empty_content_type')
def empty_content_type():
    response = Response(
            response='<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>', mimetype='')
    assert response.mimetype is None
    return response



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from django.conf.urls import url, include
except ImportError:
    try:
        from django.conf.urls.defaults import url, include
    except ImportError:
        from django.urls import re_path as url, include

from tastypie.api import Api

import views

from api import SimpleResource

simple_resource = SimpleResource()

urlpatterns = [
    url(r'^index/$', views.index, name='index'),
    url(r'^api/v1/', include(simple_resource.urls)),
]

v2_api = Api(api_name='v2')
v2_api.register(SimpleResource())

urlpatterns += [
    url(r'^api/', include(v2_api.urls)),
]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import (transient_function_wrapper,
        function_wrapper)


def validate_function_called(module, name):
    """Verify that a function is called."""

    called = []

    @transient_function_wrapper(module, name)
    def _validate_function_called(wrapped, instance, args, kwargs):
        called.append(True)
        return wrapped(*args, **kwargs)

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        new_wrapper = _validate_function_called(wrapped)
        result = new_wrapper(*args, **kwargs)
        assert called
        return result

    return wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import pytest
import six

from newrelic.api.background_task import background_task

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

from _test_common import create_request, get_result


_test_matrix = [
    ('service_method_type,service_method_method_name,raises_exception,'
    'message_count,cancel'), (
        ('unary_unary', '__call__', False, 1, False),
        ('unary_unary', '__call__', True, 1, False),
        ('unary_unary', 'with_call', False, 1, False),
        ('unary_unary', 'with_call', True, 1, False),
        ('unary_unary', 'future', False, 1, False),
        ('unary_unary', 'future', True, 1, False),
        ('unary_unary', 'future', False, 1, True),

        ('stream_unary', '__call__', False, 1, False),
        ('stream_unary', '__call__', True, 1, False),
        ('stream_unary', 'with_call', False, 1, False),
        ('stream_unary', 'with_call', True, 1, False),
        ('stream_unary', 'future', False, 1, False),
        ('stream_unary', 'future', True, 1, False),
        ('stream_unary', 'future', False, 1, True),

        ('unary_stream', '__call__', False, 1, False),
        ('unary_stream', '__call__', True, 1, False),
        ('unary_stream', '__call__', False, 2, False),
        ('unary_stream', '__call__', True, 2, False),
        ('unary_stream', '__call__', False, 1, True),
        ('unary_stream', '__call__', False, 2, True),

        ('stream_stream', '__call__', False, 1, False),
        ('stream_stream', '__call__', True, 1, False),
        ('stream_stream', '__call__', False, 2, False),
        ('stream_stream', '__call__', True, 2, False),
        ('stream_stream', '__call__', False, 1, True),
        ('stream_stream', '__call__', False, 2, True),
)]


@pytest.mark.parametrize(*_test_matrix)
def test_client(service_method_type, service_method_method_name,
        raises_exception, message_count, cancel, mock_grpc_server, stub):

    port = mock_grpc_server

    service_method_class_name = 'NoTxn%s%s' % (
            service_method_type.title().replace('_', ''),
            'Raises' if raises_exception else '')
    streaming_request = service_method_type.split('_')[0] == 'stream'
    streaming_response = service_method_type.split('_')[1] == 'stream'

    _test_scoped_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
    ]
    _test_rollup_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
            ('External/localhost:%s/all' % port, 1),
            ('External/allOther', 1),
            ('External/all', 1),
    ]

    if six.PY2:
        _test_transaction_name = 'test_clients:_test_client'
    else:
        _test_transaction_name = (
                'test_clients:test_client.<locals>._test_client')

    _errors = []
    if not streaming_response and cancel:
        _errors.append('grpc:FutureCancelledError')
    elif raises_exception or cancel:
        _errors.append('grpc._channel:_Rendezvous')

    @validate_transaction_errors(errors=_errors)
    @validate_transaction_metrics(_test_transaction_name,
            scoped_metrics=_test_scoped_metrics,
            rollup_metrics=_test_rollup_metrics,
            background_task=True)
    @background_task()
    def _test_client():
        service_method_class = getattr(stub, service_method_class_name)
        service_method_method = getattr(service_method_class,
                service_method_method_name)

        # In the case that we're preparing to cancel a request it's important
        # that the request does not return prior to cancelling. If the request
        # returns prior to cancellation then the response might be valid. In
        # order to force the request to not return, the timesout option is set.
        request = create_request(streaming_request, count=message_count,
                timesout=cancel)

        reply = service_method_method(request, None, None, None)

        if isinstance(reply, tuple):
            reply = reply[0]

        if cancel:
            reply.cancel()

        try:
            # If the reply was canceled or the server code raises an exception,
            # this will raise an exception which will be recorded by the agent
            if streaming_response:
                reply = list(reply)
            else:
                reply = [reply.result()]
        except (AttributeError, TypeError):
            reply = [reply]

        expected_text = '%s: Hello World' % service_method_type
        response_texts_correct = [r.text == expected_text for r in reply]
        assert len(response_texts_correct) == message_count
        assert response_texts_correct and all(response_texts_correct)

    try:
        _test_client()
    except grpc.RpcError as e:
        if raises_exception:
            assert '%s: Hello World' % service_method_type in e.details()
        elif cancel:
            assert e.code() == grpc.StatusCode.CANCELLED
        else:
            raise
    except grpc.FutureCancelledError:
        assert cancel


_test_matrix = [
    ('service_method_type,service_method_method_name,future_response'), (
        ('unary_unary', '__call__', False),
        ('unary_unary', 'with_call', False),
        ('unary_unary', 'future', True),

        ('stream_unary', '__call__', False),
        ('stream_unary', 'with_call', False),
        ('stream_unary', 'future', True),

        ('unary_stream', '__call__', True),

        ('stream_stream', '__call__', True),
)]


@pytest.mark.parametrize(*_test_matrix)
def test_future_timeout_error(service_method_type, service_method_method_name,
        future_response, mock_grpc_server, stub):
    port = mock_grpc_server

    service_method_class_name = 'NoTxn%s' % (
            service_method_type.title().replace('_', ''))
    streaming_request = service_method_type.split('_')[0] == 'stream'

    _test_scoped_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
    ]
    _test_rollup_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
            ('External/localhost:%s/all' % port, 1),
            ('External/allOther', 1),
            ('External/all', 1),
    ]

    if six.PY2:
        _test_transaction_name = 'test_clients:_test_future_timeout_error'
    else:
        _test_transaction_name = (
                'test_clients:test_future_timeout_error.<locals>.'
                '_test_future_timeout_error')

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(_test_transaction_name,
            scoped_metrics=_test_scoped_metrics,
            rollup_metrics=_test_rollup_metrics,
            background_task=True)
    @background_task()
    def _test_future_timeout_error():
        service_method_class = getattr(stub, service_method_class_name)
        service_method_method = getattr(service_method_class,
                service_method_method_name)

        request = create_request(streaming_request, count=1, timesout=True)

        reply = get_result(service_method_method, request, timeout=0.01)
        assert reply and reply.code() == grpc.StatusCode.DEADLINE_EXCEEDED

    _test_future_timeout_error()


_test_matrix = [
    ('service_method_type,service_method_method_name'), (
        ('unary_unary', 'with_call'),
        ('unary_unary', 'future'),

        ('stream_unary', 'with_call'),
        ('stream_unary', 'future'),

)]


@pytest.mark.parametrize(*_test_matrix)
def test_repeated_result(service_method_type, service_method_method_name,
        mock_grpc_server, stub):
    port = mock_grpc_server

    service_method_class_name = 'NoTxn%s' % (
            service_method_type.title().replace('_', ''))
    streaming_request = service_method_type.split('_')[0] == 'stream'

    _test_scoped_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
    ]
    _test_rollup_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
            ('External/localhost:%s/all' % port, 1),
            ('External/allOther', 1),
            ('External/all', 1),
    ]

    if six.PY2:
        _test_transaction_name = 'test_clients:_test_repeated_result'
    else:
        _test_transaction_name = (
                'test_clients:'
                'test_repeated_result.<locals>._test_repeated_result')

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(_test_transaction_name,
            scoped_metrics=_test_scoped_metrics,
            rollup_metrics=_test_rollup_metrics,
            background_task=True)
    @background_task()
    def _test_repeated_result():
        service_method_class = getattr(stub, service_method_class_name)
        service_method_method = getattr(service_method_class,
                service_method_method_name)

        request = create_request(streaming_request, count=1, timesout=False)

        reply = service_method_method(request)
        if isinstance(reply, tuple):
            reply = reply[1]

        reply.result()
        reply.result()

    _test_repeated_result()


_test_matrix = [
    ('service_method_type,service_method_method_name,future_response'), (
        ('unary_stream', '__call__', True),
        ('stream_stream', '__call__', True),
)]


@pytest.mark.parametrize(*_test_matrix)
def test_future_cancel(service_method_type, service_method_method_name,
        future_response, mock_grpc_server, stub):
    port = mock_grpc_server

    service_method_class_name = 'NoTxn%s' % (
            service_method_type.title().replace('_', ''))
    streaming_request = service_method_type.split('_')[0] == 'stream'

    _test_scoped_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
    ]
    _test_rollup_metrics = [
            ('External/localhost:%s/gRPC/SampleApplication/%s' % (port,
                service_method_class_name), 1),
            ('External/localhost:%s/all' % port, 1),
            ('External/allOther', 1),
            ('External/all', 1),
    ]

    if six.PY2:
        _test_transaction_name = 'test_clients:_test_future_cancel'
    else:
        _test_transaction_name = (
                'test_clients:test_future_cancel.<locals>.'
                '_test_future_cancel')

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(_test_transaction_name,
            scoped_metrics=_test_scoped_metrics,
            rollup_metrics=_test_rollup_metrics,
            background_task=True)
    @background_task()
    def _test_future_cancel():
        service_method_class = getattr(stub, service_method_class_name)
        service_method_method = getattr(service_method_class,
                service_method_method_name)

        request = create_request(streaming_request, count=3, timesout=False)

        reply = service_method_method(request)
        for result in reply:
            reply.cancel()
            break

    _test_future_cancel()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from testing_support.fixtures import override_application_settings, reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import validate_log_event_count_outside_transaction
from testing_support.validators.validate_log_events import validate_log_events
from testing_support.validators.validate_log_events_outside_transaction import validate_log_events_outside_transaction


_event_attributes = {"message": "A"}


@override_application_settings({
    "application_logging.forwarding.context_data.enabled": True,
})
def test_attributes_inside_transaction(exercise_logging_single_line):
    @validate_log_events([_event_attributes])
    @validate_log_event_count(1)
    @background_task()
    def test():
        exercise_logging_single_line()

    test()


@reset_core_stats_engine()
@override_application_settings({
    "application_logging.forwarding.context_data.enabled": True,
})
def test_attributes_outside_transaction(exercise_logging_single_line):
    @validate_log_events_outside_transaction([_event_attributes])
    @validate_log_event_count_outside_transaction(1)
    def test():
        exercise_logging_single_line()

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    from _test_not_found import _test_application
    return _test_application

_test_error_handler_not_found_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_not_found:page_not_found', 1),
        ('Function/flask.app:Flask.handle_http_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_not_found:page_not_found',
        scoped_metrics=_test_error_handler_not_found_scoped_metrics)
def test_error_handler_not_found():
    application = target_application()
    response = application.get('/missing', status=404)



# Copyright 2010 New Relic, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import isawaitable
import pytest

from newrelic.hooks.datastore_aioredis import _conn_attrs_to_dict, _instance_info
from conftest import aioredis, AIOREDIS_VERSION, SKIPIF_AIOREDIS_V1

_instance_info_tests = [
    ({}, ("localhost", "6379", "0")),
    ({"host": None}, ("localhost", "6379", "0")),
    ({"host": ""}, ("localhost", "6379", "0")),
    ({"db": None}, ("localhost", "6379", "0")),
    ({"db": ""}, ("localhost", "6379", "0")),
    ({"host": "127.0.0.1", "port": 1234, "db": 2}, ("127.0.0.1", "1234", "2")),
]


if AIOREDIS_VERSION >= (2, 0):
    clients = [aioredis.Redis, aioredis.StrictRedis]
    class DisabledConnection(aioredis.Connection):
        @staticmethod
        async def connect(*args, **kwargs):
            pass
    
        async def can_read_destructive(self, *args, **kwargs):
            return False



    class DisabledUnixConnection(aioredis.UnixDomainSocketConnection, DisabledConnection):
        pass

else:
    clients = []
    DisabledConnection, DisabledUnixConnection = None, None



@SKIPIF_AIOREDIS_V1
@pytest.mark.parametrize("client_cls", clients)
@pytest.mark.parametrize("kwargs,expected", _instance_info_tests)
def test_strict_redis_client_instance_info(client_cls, kwargs, expected, loop):
    r = client_cls(**kwargs)
    if isawaitable(r):
        r = loop.run_until_complete(r)
    conn_kwargs = r.connection_pool.connection_kwargs
    assert _instance_info(conn_kwargs) == expected


@SKIPIF_AIOREDIS_V1
@pytest.mark.parametrize("client_cls", clients)
@pytest.mark.parametrize("kwargs,expected", _instance_info_tests)
def test_strict_redis_connection_instance_info(client_cls, kwargs, expected, loop):
    r = client_cls(**kwargs)
    if isawaitable(r):
        r = loop.run_until_complete(r)
    r.connection_pool.connection_class = DisabledConnection
    connection = loop.run_until_complete(r.connection_pool.get_connection("SELECT"))
    try:
        conn_kwargs = _conn_attrs_to_dict(connection)
        assert _instance_info(conn_kwargs) == expected
    finally:
        loop.run_until_complete(r.connection_pool.release(connection))


_instance_info_from_url_tests = [
    (("redis://localhost:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://localhost:1234",), {}, ("localhost", "1234", "0")),
    (("redis://user:password@localhost:6379",), {}, ("localhost", "6379", "0")),
    (("redis://localhost:6379/2",), {}, ("localhost", "6379", "2")),
    (("redis://localhost:6379",), {"db": 2}, ("localhost", "6379", "2")),
    (("redis://@127.0.0.1:6379",), {}, ("127.0.0.1", "6379", "0")),
    (("redis://:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://@:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://localhost:1234/garbage",), {}, ("localhost", "1234", "0")),
    (("redis://127.0.0.1",), {}, ("127.0.0.1", "6379", "0")),
    (("rediss://localhost:6379/2/",), {}, ("localhost", "6379", "2")),          # rediss: Not a typo
    (("redis://localhost:6379",), {"host": "someotherhost"}, ("localhost", "6379", "0")),
    (("redis://localhost:6379/2",), {"db": 3}, ("localhost", "6379", "2")),
    (("redis://localhost:6379/2/?db=111",), {}, ("localhost", "6379", "111")),
    (("redis://localhost:6379?db=2",), {}, ("localhost", "6379", "2")),
    (("redis://localhost:6379/2?db=111",), {}, ("localhost", "6379", "111")),
    (("unix:///path/to/socket.sock",), {}, ("localhost", "/path/to/socket.sock", "0")),
    (("unix:///path/to/socket.sock?db=2",), {}, ("localhost", "/path/to/socket.sock", "2")),
    (("unix:///path/to/socket.sock",), {"db": 2}, ("localhost", "/path/to/socket.sock", "2")),
]


@SKIPIF_AIOREDIS_V1
@pytest.mark.parametrize("client_cls", clients)
@pytest.mark.parametrize("args,kwargs,expected", _instance_info_from_url_tests)
def test_strict_redis_client_from_url(client_cls, args, kwargs, expected):
    r = client_cls.from_url(*args, **kwargs)
    conn_kwargs = r.connection_pool.connection_kwargs
    assert _instance_info(conn_kwargs) == expected


@SKIPIF_AIOREDIS_V1
@pytest.mark.parametrize("client_cls", clients)
@pytest.mark.parametrize("args,kwargs,expected", _instance_info_from_url_tests)
def test_strict_redis_connection_from_url(client_cls, args, kwargs, expected, loop):
    r = client_cls.from_url(*args, **kwargs)
    if r.connection_pool.connection_class in (aioredis.Connection, aioredis.connection.SSLConnection):
        r.connection_pool.connection_class = DisabledConnection
    elif r.connection_pool.connection_class is aioredis.UnixDomainSocketConnection:
        r.connection_pool.connection_class = DisabledUnixConnection
    else:
        assert False, r.connection_pool.connection_class

    connection = loop.run_until_complete(r.connection_pool.get_connection("SELECT"))
    try:
        conn_kwargs = _conn_attrs_to_dict(connection)
        assert _instance_info(conn_kwargs) == expected
    finally:
        loop.run_until_complete(r.connection_pool.release(connection))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import psycopg2
import psycopg2.extras

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from utils import DB_SETTINGS

from newrelic.api.background_task import background_task


@validate_transaction_metrics('test_register:test_register_json',
        background_task=True)
@validate_transaction_errors(errors=[])
@background_task()
def test_register_json():
    with psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port']) as connection:

        cursor = connection.cursor()

        psycopg2.extras.register_json(connection, loads=lambda x: x)
        psycopg2.extras.register_json(cursor, loads=lambda x: x)

@validate_transaction_metrics('test_register:test_register_range',
        background_task=True)
@validate_transaction_errors(errors=[])
@background_task()
def test_register_range():
    with psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port']) as connection:

        type_name = "floatrange_" + str(os.getpid())

        create_sql = ('CREATE TYPE %s AS RANGE (' % type_name +
                      'subtype = float8,'
                      'subtype_diff = float8mi)')

        cursor = connection.cursor()

        cursor.execute("DROP TYPE if exists %s" % type_name)
        cursor.execute(create_sql)

        psycopg2.extras.register_range(type_name,
                psycopg2.extras.NumericRange, connection)

        cursor.execute("DROP TYPE if exists %s" % type_name)
        cursor.execute(create_sql)

        psycopg2.extras.register_range(type_name,
                psycopg2.extras.NumericRange, cursor)

        cursor.execute("DROP TYPE if exists %s" % type_name)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.encoding_utils import (serverless_payload_decode,
        json_decode)
from newrelic.common.object_wrapper import (
        transient_function_wrapper,
        function_wrapper)


def validate_serverless_payload(count=1):

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        payloads = []

        @transient_function_wrapper('newrelic.core.data_collector',
                'ServerlessModeSession.finalize')
        def _capture(wrapped, instance, args, kwargs):
            payload = wrapped(*args, **kwargs)
            payloads.append(payload)
            return payload

        def _validate():
            assert len(payloads) == count

            for payload in payloads:
                assert isinstance(payload, str)

                obj = json_decode(payload)

                assert len(obj) == 3, obj

                assert obj[0] == 1  # Version = 1
                assert obj[1] == 'NR_LAMBDA_MONITORING'  # Marker

                decoded = serverless_payload_decode(obj[2])

                # Keys should only contain metadata / data
                set(decoded.keys()) == set(('metadata', 'data'))

        capture_wrapped = _capture(wrapped)
        result = capture_wrapped(*args, **kwargs)
        _validate()
        return result

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.common.encoding_utils import camel_case, snake_case


@pytest.mark.parametrize("input_,expected,upper", [
    ("", "", False),
    ("", "", True),
    ("my_string", "myString", False),
    ("my_string", "MyString", True),
    ("LeaveCase", "LeaveCase", False),
    ("correctCase", "CorrectCase", True),
    ("UPPERcaseLETTERS", "UPPERcaseLETTERS", False),
    ("UPPERcaseLETTERS", "UPPERcaseLETTERS", True),
    ("lowerCASEletters", "lowerCASEletters", False),
    ("lowerCASEletters", "LowerCASEletters", True),
    ("very_long_snake_string", "VeryLongSnakeString", True),
    ("kebab-case", "kebab-case", False),
])
def test_camel_case(input_, expected, upper):
    output = camel_case(input_, upper=upper)
    assert output == expected


@pytest.mark.parametrize("input_,expected", [
    ("", ""),
    ("", ""),
    ("my_string", "my_string"),
    ("myString", "my_string"),
    ("MyString", "my_string"),
    ("UPPERcaseLETTERS", "uppercase_letters"),
    ("lowerCASEletters", "lower_caseletters"),
    ("VeryLongCamelString", "very_long_camel_string"),
    ("kebab-case", "kebab-case"),
])
def test_snake_case(input_, expected):
    output = snake_case(input_)
    assert output == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import function_wrapper


def validate_distributed_trace_accepted(header="newrelic", transport_type="HTTP"):
    @function_wrapper
    def _validate_distributed_trace_accepted(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        txn = current_transaction()

        assert txn
        assert txn._distributed_trace_state
        assert txn.parent_type == "App"
        assert txn._trace_id.startswith(txn.parent_tx)
        assert txn.parent_span is not None
        assert txn.parent_account == txn.settings.account_id
        assert txn.parent_transport_type == transport_type
        assert txn._priority is not None
        assert txn._sampled is not None

        return result

    return _validate_distributed_trace_accepted



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import memcache

from newrelic.api.transaction import current_transaction
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.db_settings import memcached_settings
from testing_support.util import instance_hostname

from newrelic.api.background_task import background_task

DB_SETTINGS = memcached_settings()[0]
MEMCACHED_ADDR = '%s:%s' % (DB_SETTINGS['host'], DB_SETTINGS['port'])

# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
    'distributed_tracing.enabled': True,
    'span_events.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
    'distributed_tracing.enabled': True,
    'span_events.enabled': True,
}


# Query

def _exercise_db(client):
    key = DB_SETTINGS['namespace'] + 'key'
    client.set(key, 'value')
    value = client.get(key)
    client.delete(key)
    assert value == 'value'


# Tests

@pytest.mark.parametrize('instance_enabled', (True, False))
def test_span_events(instance_enabled):
    guid = 'dbb533c53b749e0b'
    priority = 0.5

    common = {
        'type': 'Span',
        'transactionId': guid,
        'priority': priority,
        'sampled': True,
        'category': 'datastore',
        'component': 'Memcached',
        'span.kind': 'client',
    }

    exact_agents = {}
    if instance_enabled:
        settings = _enable_instance_settings
        hostname = instance_hostname(DB_SETTINGS['host'])
        exact_agents.update({
            'peer.address': '%s:%s' % (hostname, DB_SETTINGS['port']),
            'peer.hostname': hostname,
        })
    else:
        settings = _disable_instance_settings
        exact_agents.update({
            'peer.address': 'Unknown:Unknown',
            'peer.hostname': 'Unknown',
        })

    query_1 = common.copy()
    query_1['name'] = 'Datastore/operation/Memcached/set'

    query_2 = common.copy()
    query_2['name'] = 'Datastore/operation/Memcached/get'

    query_3 = common.copy()
    query_3['name'] = 'Datastore/operation/Memcached/delete'

    @validate_span_events(count=1, exact_intrinsics=query_1,
            exact_agents=exact_agents,
            unexpected_agents=('db.instance',))
    @validate_span_events(count=1, exact_intrinsics=query_2,
            exact_agents=exact_agents,
            unexpected_agents=('db.instance',))
    @validate_span_events(count=1, exact_intrinsics=query_3,
            exact_agents=exact_agents,
            unexpected_agents=('db.instance',))
    @validate_span_events(count=0, expected_intrinsics=('db.statement',))
    @override_application_settings(settings)
    @background_task(name='span_events')
    def _test():
        txn = current_transaction()
        txn.guid = guid
        txn._priority = priority
        txn._sampled = True

        client = memcache.Client([MEMCACHED_ADDR])
        _exercise_db(client)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains a special case test for CAT. Most CAT testing is found at
`tests/cross_agent/test_cat_map.py` and
`newrelic/api/tests/test_cross_process.py`. This test does not fit either of
those spaces, the former being reserved for cross agent testing and the latter
being a unittest for the `process_response` method. Since this is a more end to
end style test, it does not fit as a unittest.
"""

import pytest
import webtest
from testing_support.fixtures import (
    cat_enabled,
    make_cross_agent_headers,
    override_application_settings,
)

from newrelic.api.background_task import background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.wsgi_application import wsgi_application

ENCODING_KEY = "1234567890123456789012345678901234567890"


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status_code = int(environ["PATH_INFO"].strip("/"))
    status = "%d STATUS" % status_code

    if status_code == 304:
        output = b""
        response_headers = []
    else:
        output = b"hello world"
        response_headers = [("Content-type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


test_application = webtest.TestApp(target_wsgi_application)

_override_settings = {
    "cross_process_id": "1#1",
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [1],
    "browser_monitoring.enabled": False,
}

payload = ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"]


@cat_enabled
@override_application_settings(_override_settings)
def test_cat_disabled_browser_monitoring():
    headers = make_cross_agent_headers(payload, ENCODING_KEY, "1#1")
    response = test_application.get("/200", headers=headers)
    assert "X-NewRelic-App-Data" in response.headers


@override_application_settings(_override_settings)
def test_cat_insertion_disabled_on_304():
    headers = make_cross_agent_headers(payload, ENCODING_KEY, "1#1")
    response = test_application.get("/304", headers=headers)
    assert "X-NewRelic-App-Data" not in response.headers


_override_settings = {
    "cross_application_tracing.enabled": True,
    "distributed_tracing.enabled": False,
}


@cat_enabled
@override_application_settings(_override_settings)
@pytest.mark.parametrize("fips_enabled", (False, True))
@background_task()
def test_cat_fips_compliance(monkeypatch, fips_enabled):
    # Set md5 to raise a ValueError to simulate FIPS compliance issues.
    def md5_crash(*args, **kwargs):
        raise ValueError()

    if fips_enabled:
        # monkeypatch.setattr("hashlib.md5", md5_crash)
        import hashlib

        monkeypatch.setattr(hashlib, "md5", md5_crash)

    # Generate and send request using actual transaction api instead of fixture.
    # Otherwise the proper code paths are not exercised.
    with ExternalTrace("cat_test", "http://localhost/200") as tracer:
        headers = tracer.generate_request_headers(tracer.transaction)

    expected = not fips_enabled  # Invert to make more human readable
    assert ("X-NewRelic-Transaction" in dict(headers)) == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import pytest

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import register_database_client
from newrelic.api.transaction import current_transaction
from newrelic.core.database_node import DatabaseNode
from newrelic.core.stats_engine import StatsEngine

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.join(CURRENT_DIR, "fixtures", "datastores", "datastore_instances.json")

_parameters_list = [
    "name",
    "system_hostname",
    "db_hostname",
    "product",
    "port",
    "unix_socket",
    "database_path",
    "expected_instance_metric",
]

_parameters = ",".join(_parameters_list)


def _load_tests():
    with open(FIXTURE, "r") as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None if f != "db_hostname" else "localhost") for f in _parameters_list])


_datastore_tests = [_parametrize_test(t) for t in _load_tests()]


@pytest.mark.parametrize(_parameters, _datastore_tests)
@background_task()
def test_datastore_instance(
    name, system_hostname, db_hostname, product, port, unix_socket, database_path, expected_instance_metric, monkeypatch
):

    monkeypatch.setattr("newrelic.common.system_info.gethostname", lambda: system_hostname)

    class FakeModule:
        pass

    register_database_client(FakeModule, product)

    port_path_or_id = port or database_path or unix_socket

    node = DatabaseNode(
        dbapi2_module=FakeModule,
        sql="",
        children=[],
        start_time=0,
        end_time=1,
        duration=1,
        exclusive=1,
        stack_trace=None,
        sql_format="obfuscated",
        connect_params=None,
        cursor_params=None,
        sql_parameters=None,
        execute_params=None,
        host=db_hostname,
        port_path_or_id=port_path_or_id,
        database_name=database_path,
        guid=None,
        agent_attributes={},
        user_attributes={},
    )

    empty_stats = StatsEngine()
    transaction = current_transaction()
    unscoped_scope = ""

    # Check 'Datastore/instance' metric to confirm that:
    #   1. metric name is reported correctly
    #   2. metric scope is always unscoped

    for metric in node.time_metrics(empty_stats, transaction, None):
        if metric.name.startswith("Datastore/instance/"):
            assert metric.name == expected_instance_metric, name
            assert metric.scope == unscoped_scope



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from kafka.serializer import Serializer

from newrelic.api.application import application_instance
from newrelic.api.function_trace import FunctionTraceWrapper
from newrelic.api.message_trace import MessageTrace
from newrelic.api.message_transaction import MessageTransaction
from newrelic.api.time_trace import current_trace, notice_error
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import (
    ObjectProxy,
    function_wrapper,
    wrap_function_wrapper,
)
from newrelic.common.package_version_utils import get_package_version

HEARTBEAT_POLL = "MessageBroker/Kafka/Heartbeat/Poll"
HEARTBEAT_SENT = "MessageBroker/Kafka/Heartbeat/Sent"
HEARTBEAT_FAIL = "MessageBroker/Kafka/Heartbeat/Fail"
HEARTBEAT_RECEIVE = "MessageBroker/Kafka/Heartbeat/Receive"
HEARTBEAT_SESSION_TIMEOUT = "MessageBroker/Kafka/Heartbeat/SessionTimeout"
HEARTBEAT_POLL_TIMEOUT = "MessageBroker/Kafka/Heartbeat/PollTimeout"


def _bind_send(topic, value=None, key=None, headers=None, partition=None, timestamp_ms=None):
    return topic, value, key, headers, partition, timestamp_ms


def wrap_KafkaProducer_send(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    topic, value, key, headers, partition, timestamp_ms = _bind_send(*args, **kwargs)
    headers = list(headers) if headers else []

    transaction.add_messagebroker_info("Kafka-Python", get_package_version("kafka-python"))

    with MessageTrace(
        library="Kafka",
        operation="Produce",
        destination_type="Topic",
        destination_name=topic or "Default",
        source=wrapped,
        terminal=False,
    ) as trace:
        dt_headers = [(k, v.encode("utf-8")) for k, v in trace.generate_request_headers(transaction)]
        headers.extend(dt_headers)
        try:
            return wrapped(topic, value=value, key=key, headers=headers, partition=partition, timestamp_ms=timestamp_ms)
        except Exception:
            notice_error()
            raise


def wrap_kafkaconsumer_next(wrapped, instance, args, kwargs):
    if hasattr(instance, "_nr_transaction") and not instance._nr_transaction.stopped:
        instance._nr_transaction.__exit__(*sys.exc_info())

    try:
        record = wrapped(*args, **kwargs)
    except Exception as e:
        # StopIteration is an expected error, indicating the end of an iterable,
        # that should not be captured.
        if not isinstance(e, StopIteration):
            if current_transaction():
                # Report error on existing transaction if there is one
                notice_error()
            else:
                # Report error on application
                notice_error(application=application_instance(activate=False))
        raise

    if record:
        # This iterator can be called either outside of a transaction, or
        # within the context of an existing transaction.  There are 3
        # possibilities we need to handle: (Note that this is similar to
        # our Pika and Celery instrumentation)
        #
        #   1. In an inactive transaction
        #
        #      If the end_of_transaction() or ignore_transaction() API
        #      calls have been invoked, this iterator may be called in the
        #      context of an inactive transaction. In this case, don't wrap
        #      the iterator in any way. Just run the original iterator.
        #
        #   2. In an active transaction
        #
        #      Do nothing.
        #
        #   3. Outside of a transaction
        #
        #      Since it's not running inside of an existing transaction, we
        #      want to create a new background transaction for it.

        library = "Kafka"
        destination_type = "Topic"
        destination_name = record.topic
        received_bytes = len(str(record.value).encode("utf-8"))
        message_count = 1

        transaction = current_transaction(active_only=False)

        if not transaction:
            transaction = MessageTransaction(
                application=application_instance(),
                library=library,
                destination_type=destination_type,
                destination_name=destination_name,
                headers=dict(record.headers),
                transport_type="Kafka",
                routing_key=record.key,
                source=wrapped,
            )
            instance._nr_transaction = transaction
            transaction.__enter__()  # pylint: disable=C2801

            # Obtain consumer client_id to send up as agent attribute
            if hasattr(instance, "config") and "client_id" in instance.config:
                client_id = instance.config["client_id"]
                transaction._add_agent_attribute("kafka.consume.client_id", client_id)

            transaction._add_agent_attribute("kafka.consume.byteCount", received_bytes)

        transaction = current_transaction()
        if transaction:  # If there is an active transaction now.
            # Add metrics whether or not a transaction was already active, or one was just started.
            # Don't add metrics if there was an inactive transaction.
            # Name the metrics using the same format as the transaction, but in case the active transaction
            # was an existing one and not a message transaction, reproduce the naming logic here.
            group = "Message/%s/%s" % (library, destination_type)
            name = "Named/%s" % destination_name
            transaction.record_custom_metric("%s/%s/Received/Bytes" % (group, name), received_bytes)
            transaction.record_custom_metric("%s/%s/Received/Messages" % (group, name), message_count)
            transaction.add_messagebroker_info("Kafka-Python", get_package_version("kafka-python"))

    return record


def wrap_KafkaProducer_init(wrapped, instance, args, kwargs):
    get_config_key = lambda key: kwargs.get(key, instance.DEFAULT_CONFIG[key])  # pylint: disable=C3001 # noqa: E731

    kwargs["key_serializer"] = wrap_serializer(
        instance, "Serialization/Key", "MessageBroker", get_config_key("key_serializer")
    )
    kwargs["value_serializer"] = wrap_serializer(
        instance, "Serialization/Value", "MessageBroker", get_config_key("value_serializer")
    )

    return wrapped(*args, **kwargs)


class NewRelicSerializerWrapper(ObjectProxy):
    def __init__(self, wrapped, serializer_name, group_prefix):
        ObjectProxy.__init__.__get__(self)(wrapped)  # pylint: disable=W0231

        self._nr_serializer_name = serializer_name
        self._nr_group_prefix = group_prefix

    def serialize(self, topic, object):
        wrapped = self.__wrapped__.serialize  # pylint: disable=W0622
        args = (topic, object)
        kwargs = {}

        if not current_transaction():
            return wrapped(*args, **kwargs)

        group = "%s/Kafka/Topic" % self._nr_group_prefix
        name = "Named/%s/%s" % (topic, self._nr_serializer_name)

        return FunctionTraceWrapper(wrapped, name=name, group=group)(*args, **kwargs)


def wrap_serializer(client, serializer_name, group_prefix, serializer):
    @function_wrapper
    def _wrap_serializer(wrapped, instance, args, kwargs):
        transaction = current_transaction()
        if not transaction:
            return wrapped(*args, **kwargs)

        topic = "Unknown"
        if isinstance(transaction, MessageTransaction):
            topic = transaction.destination_name
        else:
            # Find parent message trace to retrieve topic
            message_trace = current_trace()
            while message_trace is not None and not isinstance(message_trace, MessageTrace):
                message_trace = message_trace.parent
            if message_trace:
                topic = message_trace.destination_name

        group = "%s/Kafka/Topic" % group_prefix
        name = "Named/%s/%s" % (topic, serializer_name)

        return FunctionTraceWrapper(wrapped, name=name, group=group)(*args, **kwargs)

    try:
        # Apply wrapper to serializer
        if serializer is None:
            # Do nothing
            return serializer
        elif isinstance(serializer, Serializer):
            return NewRelicSerializerWrapper(serializer, group_prefix=group_prefix, serializer_name=serializer_name)
        else:
            # Wrap callable in wrapper
            return _wrap_serializer(serializer)
    except Exception:
        return serializer  # Avoid crashes from immutable serializers


def metric_wrapper(metric_name, check_result=False):
    def _metric_wrapper(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        application = application_instance(activate=False)
        if application:
            if not check_result or check_result and result:
                # If the result does not need validated, send metric.
                # If the result does need validated, ensure it is True.
                application.record_custom_metric(metric_name, 1)

        return result

    return _metric_wrapper


def instrument_kafka_producer(module):
    if hasattr(module, "KafkaProducer"):
        wrap_function_wrapper(module, "KafkaProducer.__init__", wrap_KafkaProducer_init)
        wrap_function_wrapper(module, "KafkaProducer.send", wrap_KafkaProducer_send)


def instrument_kafka_consumer_group(module):
    if hasattr(module, "KafkaConsumer"):
        wrap_function_wrapper(module, "KafkaConsumer.__next__", wrap_kafkaconsumer_next)


def instrument_kafka_heartbeat(module):
    if hasattr(module, "Heartbeat"):
        if hasattr(module.Heartbeat, "poll"):
            wrap_function_wrapper(module, "Heartbeat.poll", metric_wrapper(HEARTBEAT_POLL))

        if hasattr(module.Heartbeat, "fail_heartbeat"):
            wrap_function_wrapper(module, "Heartbeat.fail_heartbeat", metric_wrapper(HEARTBEAT_FAIL))

        if hasattr(module.Heartbeat, "sent_heartbeat"):
            wrap_function_wrapper(module, "Heartbeat.sent_heartbeat", metric_wrapper(HEARTBEAT_SENT))

        if hasattr(module.Heartbeat, "received_heartbeat"):
            wrap_function_wrapper(module, "Heartbeat.received_heartbeat", metric_wrapper(HEARTBEAT_RECEIVE))

        if hasattr(module.Heartbeat, "session_timeout_expired"):
            wrap_function_wrapper(
                module,
                "Heartbeat.session_timeout_expired",
                metric_wrapper(HEARTBEAT_SESSION_TIMEOUT, check_result=True),
            )

        if hasattr(module.Heartbeat, "poll_timeout_expired"):
            wrap_function_wrapper(
                module, "Heartbeat.poll_timeout_expired", metric_wrapper(HEARTBEAT_POLL_TIMEOUT, check_result=True)
            )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

try:
    import http.client as httplib
except ImportError:
    import httplib

from testing_support.external_fixtures import (
    cache_outgoing_headers,
    insert_incoming_headers,
)
from testing_support.fixtures import cat_enabled, override_application_settings
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_external_node_params import (
    validate_external_node_params,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.common.encoding_utils import DistributedTracePayload
from newrelic.packages import six


def select_python_version(py2, py3):
    return six.PY3 and py3 or py2


def test_httplib_http_request(server):
    scoped = [
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        )
    ]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        ),
    ]

    @validate_transaction_metrics(
        "test_httplib:test_httplib_http_request", scoped_metrics=scoped, rollup_metrics=rollup, background_task=True
    )
    @background_task(name="test_httplib:test_httplib_http_request")
    def _test():
        connection = httplib.HTTPConnection("localhost", server.port)
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()
        connection.close()

    _test()


def test_httplib_https_request(server):
    _test_httplib_https_request_scoped_metrics = [
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        )
    ]

    _test_httplib_https_request_rollup_metrics = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        ),
    ]

    @validate_transaction_metrics(
        "test_httplib:test_httplib_https_request",
        scoped_metrics=_test_httplib_https_request_scoped_metrics,
        rollup_metrics=_test_httplib_https_request_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_httplib:test_httplib_https_request")
    def _test():
        # fix HTTPSConnection: https://wiki.openstack.org/wiki/OSSN/OSSN-0033
        connection = httplib.HTTPSConnection("localhost", server.port)  # nosec
        # It doesn't matter that a SSL exception is raised here because the
        # agent still records this as an external request
        try:
            connection.request("GET", "/")
        except Exception:
            pass
        connection.close()

    _test()


def test_httplib_http_with_port_request(server):

    scoped = [
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        )
    ]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        select_python_version(
            py2=("External/localhost:%d/httplib/" % server.port, 1),
            py3=("External/localhost:%d/http/" % server.port, 1),
        ),
    ]

    @validate_transaction_metrics(
        "test_httplib:test_httplib_http_with_port_request",
        scoped_metrics=scoped,
        rollup_metrics=rollup,
        background_task=True,
    )
    @background_task(name="test_httplib:test_httplib_http_with_port_request")
    def _test():
        connection = httplib.HTTPConnection("localhost", server.port)
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()
        connection.close()

    _test()


@pytest.mark.parametrize(
    "distributed_tracing,span_events",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
def test_httplib_cross_process_request(server, distributed_tracing, span_events):
    @background_task(name="test_httplib:test_httplib_cross_process_request")
    @cache_outgoing_headers
    @validate_cross_process_headers
    def _test():
        connection = httplib.HTTPConnection("localhost", server.port)
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()
        connection.close()

    _test = override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "cross_application_tracer.enabled": not distributed_tracing,
            "span_events.enabled": span_events,
        }
    )(_test)

    _test()


_test_httplib_cross_process_response_external_node_params = [
    ("cross_process_id", "1#2"),
    ("external_txn_name", "test"),
    ("transaction_guid", "0123456789012345"),
]


@cat_enabled
@insert_incoming_headers
def test_httplib_cross_process_response(server):
    scoped = [("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("ExternalApp/localhost:%d/1#2/all" % server.port, 1),
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1),
    ]

    @validate_transaction_metrics(
        "test_httplib:test_httplib_cross_process_response",
        scoped_metrics=scoped,
        rollup_metrics=rollup,
        background_task=True,
    )
    @validate_external_node_params(params=_test_httplib_cross_process_response_external_node_params)
    @background_task(name="test_httplib:test_httplib_cross_process_response")
    def _test():
        connection = httplib.HTTPConnection("localhost", server.port)
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()
        connection.close()

    _test()


@cat_enabled
def test_httplib_multiple_requests_cross_process_response(server):
    connection = httplib.HTTPConnection("localhost", server.port)

    scoped = [("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("ExternalApp/localhost:%d/1#2/all" % server.port, 1),
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1),
    ]

    @validate_transaction_metrics(
        "test_httplib:test_transaction", scoped_metrics=scoped, rollup_metrics=rollup, background_task=True
    )
    @insert_incoming_headers
    @validate_external_node_params(params=_test_httplib_cross_process_response_external_node_params)
    @background_task(name="test_httplib:test_transaction")
    def test_transaction():
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()

    # make multiple requests with the same connection
    for _ in range(2):
        test_transaction()

    connection.close()


def process_response(response):
    response = response.decode("utf-8").strip()
    values = response.splitlines()
    values = [[x.strip() for x in s.split(":", 1)] for s in values]
    return {v[0]: v[1] for v in values}


def test_httplib_multiple_requests_unique_distributed_tracing_id(server):
    connection = httplib.HTTPConnection("localhost", server.port)
    response_headers = []

    @background_task(name="test_httplib:test_transaction")
    def test_transaction():
        connection.request("GET", "/")
        response = connection.getresponse()
        response_headers.append(process_response(response.read()))
        connection.request("GET", "/")
        response = connection.getresponse()
        response_headers.append(process_response(response.read()))

    test_transaction = override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
        }
    )(test_transaction)
    # make multiple requests with the same connection
    test_transaction()

    connection.close()
    dt_payloads = [DistributedTracePayload.from_http_safe(header["newrelic"]) for header in response_headers]

    ids = set()
    for payload in dt_payloads:
        assert payload["d"]["id"] not in ids
        ids.add(payload["d"]["id"])


def test_httplib_nr_headers_added(server):
    connection = httplib.HTTPConnection("localhost", server.port)
    key = "newrelic"
    value = "testval"
    headers = []

    @background_task(name="test_httplib:test_transaction")
    def test_transaction():
        connection.putrequest("GET", "/")
        connection.putheader(key, value)
        connection.endheaders()
        response = connection.getresponse()
        headers.append(process_response(response.read()))

    test_transaction = override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
        }
    )(test_transaction)
    test_transaction()
    connection.close()
    # verify newrelic headers already added do not get overrode
    assert headers[0][key] == value


def test_span_events(server):
    connection = httplib.HTTPConnection("localhost", server.port)

    _settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }

    uri = "http://localhost:%d" % server.port

    exact_intrinsics = {
        "name": select_python_version(
            py2="External/localhost:%d/httplib/" % server.port, py3="External/localhost:%d/http/" % server.port
        ),
        "type": "Span",
        "sampled": True,
        "category": "http",
        "span.kind": "client",
        "component": select_python_version(py2="httplib", py3="http"),
    }
    exact_agents = {
        "http.url": uri,
        "http.statusCode": 200,
    }

    expected_intrinsics = ("timestamp", "duration", "transactionId")

    @override_application_settings(_settings)
    @validate_span_events(
        count=1, exact_intrinsics=exact_intrinsics, exact_agents=exact_agents, expected_intrinsics=expected_intrinsics
    )
    @validate_tt_segment_params(exact_params=exact_agents)
    @background_task(name="test_httplib:test_span_events")
    def _test():
        connection.request("GET", "/")
        response = connection.getresponse()
        response.read()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from testing_support.fixtures import (
    override_application_settings,
    raise_background_exceptions,
    wait_for_background_threads,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.package_version_utils import get_package_version

WAITRESS_VERSION = get_package_version("waitress")


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_wsgi_application_index(target_application):
    @validate_transaction_metrics(
        "_application:sample_application",
        custom_metrics=[
            ("Python/Dispatcher/Waitress/%s" % WAITRESS_VERSION, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        response = target_application.get("/")
        assert response.status == "200 OK"

    _test()


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_raise_exception_application(target_application):
    @validate_transaction_errors(["builtins:RuntimeError"])
    @validate_transaction_metrics(
        "_application:sample_application",
        custom_metrics=[
            ("Python/Dispatcher/Waitress/%s" % WAITRESS_VERSION, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        response = target_application.get("/raise-exception-application/", status=500)
        assert response.status == "500 Internal Server Error"

    _test()


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_raise_exception_response(target_application):
    @validate_transaction_errors(["builtins:RuntimeError"])
    @validate_transaction_metrics(
        "_application:sample_application",
        custom_metrics=[
            ("Python/Dispatcher/Waitress/%s" % WAITRESS_VERSION, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        response = target_application.get("/raise-exception-response/", status=500)
        assert response.status == "500 Internal Server Error"

    _test()


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_raise_exception_finalize(target_application):
    @validate_transaction_errors(["builtins:RuntimeError"])
    @validate_transaction_metrics(
        "_application:sample_application",
        custom_metrics=[
            ("Python/Dispatcher/Waitress/%s" % WAITRESS_VERSION, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        response = target_application.get("/raise-exception-finalize/", status=500)
        assert response.status == "500 Internal Server Error"

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import json

import pytest
from framework_strawberry._target_schema_async import (
    target_asgi_application as target_asgi_application_async,
)
from framework_strawberry._target_schema_async import (
    target_schema as target_schema_async,
)
from framework_strawberry._target_schema_sync import (
    target_asgi_application as target_asgi_application_sync,
)
from framework_strawberry._target_schema_sync import target_schema as target_schema_sync


def run_sync(schema):
    def _run_sync(query, middleware=None):
        from graphql.language.source import Source

        if middleware is not None:
            pytest.skip("Middleware not supported in Strawberry.")

        response = schema.execute_sync(query)

        if isinstance(query, str) and "error" not in query or isinstance(query, Source) and "error" not in query.body:
            assert not response.errors
        else:
            assert response.errors

        return response.data

    return _run_sync


def run_async(schema):
    def _run_async(query, middleware=None):
        from graphql.language.source import Source

        if middleware is not None:
            pytest.skip("Middleware not supported in Strawberry.")

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(schema.execute(query))

        if isinstance(query, str) and "error" not in query or isinstance(query, Source) and "error" not in query.body:
            assert not response.errors
        else:
            assert response.errors

        return response.data

    return _run_async


def run_asgi(app):
    def _run_asgi(query, middleware=None):
        if middleware is not None:
            pytest.skip("Middleware not supported in Strawberry.")

        response = app.make_request(
            "POST", "/", body=json.dumps({"query": query}), headers={"Content-Type": "application/json"}
        )
        body = json.loads(response.body.decode("utf-8"))

        if not isinstance(query, str) or "error" in query:
            try:
                assert response.status != 200
            except AssertionError:
                assert body["errors"]
        else:
            assert response.status == 200
            assert "errors" not in body or not body["errors"]

        return body["data"]

    return _run_asgi


target_application = {
    "sync-sync": run_sync(target_schema_sync),
    "async-sync": run_async(target_schema_sync),
    "asgi-sync": run_asgi(target_asgi_application_sync),
    "async-async": run_async(target_schema_async),
    "asgi-async": run_asgi(target_asgi_application_async),
}



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import capture_transaction_metrics

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import function_trace
from newrelic.common.object_wrapper import function_wrapper

asyncio = pytest.importorskip("asyncio")


def validate_total_time_value_greater_than(value, concurrent=False):
    @function_wrapper
    def _validate_total_time_value(wrapped, instance, args, kwargs):
        metrics = {}
        result = capture_transaction_metrics([], metrics)(wrapped)(*args, **kwargs)
        total_time = metrics[("OtherTransactionTotalTime", "")][1]
        # Assert total call time is at least that value
        assert total_time >= value

        duration = metrics[("OtherTransaction/all", "")][1]
        if concurrent:
            # If there is concurrent work, the total_time must be strictly
            # greater than the duration
            assert total_time > duration
        else:
            assert total_time == duration
        return result

    return _validate_total_time_value


@function_trace(name="child")
async def child():
    await asyncio.sleep(0.1)


@background_task(name="parent")
async def parent(calls):
    coros = [child() for _ in range(calls)]
    await asyncio.gather(*coros)
    await asyncio.sleep(0.1)


@validate_total_time_value_greater_than(0.2)
def test_total_time_sync(event_loop):
    event_loop.run_until_complete(parent(1))


@validate_total_time_value_greater_than(0.3, concurrent=True)
def test_total_time_async(event_loop):
    event_loop.run_until_complete(parent(2))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest
from testing_support.external_fixtures import validate_synthetics_external_trace_header
from testing_support.fixtures import (
    cat_enabled,
    make_synthetics_headers,
    override_application_settings,
)
from testing_support.validators.validate_synthetics_event import (
    validate_synthetics_event,
)
from testing_support.validators.validate_synthetics_transaction_trace import (
    validate_synthetics_transaction_trace,
)

from newrelic.api.web_transaction import web_transaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import deobfuscate, json_decode
from newrelic.core.agent import agent_instance

ENCODING_KEY = "1234567890123456789012345678901234567890"
ACCOUNT_ID = "444"
SYNTHETICS_RESOURCE_ID = "09845779-16ef-4fa7-b7f2-44da8e62931c"
SYNTHETICS_JOB_ID = "8c7dd3ba-4933-4cbb-b1ed-b62f511782f4"
SYNTHETICS_MONITOR_ID = "dc452ae9-1a93-4ab5-8a33-600521e9cd00"
SYNTHETICS_TYPE = "scheduled"
SYNTHETICS_INITIATOR = "graphql"
SYNTHETICS_ATTRIBUTES = {"exampleAttribute": "1"}

_override_settings = {
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [int(ACCOUNT_ID)],
    "synthetics.enabled": True,
}


def _make_synthetics_headers(
    version="1",
    account_id=ACCOUNT_ID,
    resource_id=SYNTHETICS_RESOURCE_ID,
    job_id=SYNTHETICS_JOB_ID,
    monitor_id=SYNTHETICS_MONITOR_ID,
    encoding_key=ENCODING_KEY,
    info_version="1",
    type_=SYNTHETICS_TYPE,
    initiator=SYNTHETICS_INITIATOR,
    attributes=SYNTHETICS_ATTRIBUTES,
):
    return make_synthetics_headers(encoding_key, account_id, resource_id, job_id, monitor_id, type_, initiator, attributes, synthetics_version=version, synthetics_info_version=info_version)


def decode_header(header, encoding_key=ENCODING_KEY):
    result = deobfuscate(header, encoding_key)
    return json_decode(result)


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"

    output = "<html><head>header</head><body><p>RESPONSE</p></body></html>"
    output = output.encode("UTF-8")

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


target_application = webtest.TestApp(target_wsgi_application)

_test_valid_synthetics_event_required = [
    ("nr.syntheticsResourceId", SYNTHETICS_RESOURCE_ID),
    ("nr.syntheticsJobId", SYNTHETICS_JOB_ID),
    ("nr.syntheticsMonitorId", SYNTHETICS_MONITOR_ID),
    ("nr.syntheticsType", SYNTHETICS_TYPE),
    ("nr.syntheticsInitiator", SYNTHETICS_INITIATOR),
    ("nr.syntheticsExampleAttribute", "1"),
]
_test_valid_synthetics_event_forgone = []


@validate_synthetics_event(
    _test_valid_synthetics_event_required, _test_valid_synthetics_event_forgone, should_exist=True
)
@override_application_settings(_override_settings)
def test_valid_synthetics_event():
    headers = _make_synthetics_headers()
    response = target_application.get("/", headers=headers)


_test_valid_synthetics_event_without_info_required = [
    ("nr.syntheticsResourceId", SYNTHETICS_RESOURCE_ID),
    ("nr.syntheticsJobId", SYNTHETICS_JOB_ID),
    ("nr.syntheticsMonitorId", SYNTHETICS_MONITOR_ID),
]
_test_valid_synthetics_event_without_info_forgone = [
    "nr.syntheticsType",
    "nr.syntheticsInitiator",
    "nr.syntheticsExampleAttribute",
]


@validate_synthetics_event(
    _test_valid_synthetics_event_without_info_required, _test_valid_synthetics_event_without_info_forgone, should_exist=True
)
@override_application_settings(_override_settings)
def test_valid_synthetics_event_without_info():
    headers = _make_synthetics_headers(type_=None, initiator=None, attributes=None)
    response = target_application.get("/", headers=headers)


@validate_synthetics_event([], [], should_exist=False)
@override_application_settings(_override_settings)
def test_no_synthetics_event_unsupported_version():
    headers = _make_synthetics_headers(version="0")
    response = target_application.get("/", headers=headers)


@validate_synthetics_event(
    _test_valid_synthetics_event_without_info_required, _test_valid_synthetics_event_without_info_forgone, should_exist=True
)
@override_application_settings(_override_settings)
def test_synthetics_event_unsupported_info_version():
    headers = _make_synthetics_headers(info_version="0")
    response = target_application.get("/", headers=headers)


@validate_synthetics_event([], [], should_exist=False)
@override_application_settings(_override_settings)
def test_no_synthetics_event_untrusted_account():
    headers = _make_synthetics_headers(account_id="999")
    response = target_application.get("/", headers=headers)


@validate_synthetics_event([], [], should_exist=False)
@override_application_settings(_override_settings)
def test_no_synthetics_event_mismatched_encoding_key():
    encoding_key = "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
    headers = _make_synthetics_headers(encoding_key=encoding_key)
    response = target_application.get("/", headers=headers)


@validate_synthetics_event(
    _test_valid_synthetics_event_without_info_required, _test_valid_synthetics_event_without_info_forgone, should_exist=True
)
@override_application_settings(_override_settings)
def test_synthetics_event_mismatched_info_encoding_key():
    encoding_key = "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
    headers = {
        "X-NewRelic-Synthetics": _make_synthetics_headers(type_=None)["X-NewRelic-Synthetics"],
        "X-NewRelic-Synthetics-Info": _make_synthetics_headers(encoding_key=encoding_key)["X-NewRelic-Synthetics-Info"],
    }
    response = target_application.get("/", headers=headers)


_test_valid_synthetics_tt_required = {
    "synthetics_resource_id": SYNTHETICS_RESOURCE_ID,
    "synthetics_job_id": SYNTHETICS_JOB_ID,
    "synthetics_monitor_id": SYNTHETICS_MONITOR_ID,
    "synthetics_type": SYNTHETICS_TYPE,
    "synthetics_initiator": SYNTHETICS_INITIATOR,
    "synthetics_example_attribute": "1",
}


@cat_enabled
@validate_synthetics_transaction_trace(_test_valid_synthetics_tt_required)
@override_application_settings(_override_settings)
def test_valid_synthetics_in_transaction_trace():
    headers = _make_synthetics_headers()
    response = target_application.get("/", headers=headers)


@validate_synthetics_transaction_trace([], _test_valid_synthetics_tt_required, should_exist=False)
@override_application_settings(_override_settings)
def test_no_synthetics_in_transaction_trace():
    response = target_application.get("/")


_disabled_settings = {
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [int(ACCOUNT_ID)],
    "synthetics.enabled": False,
}


@validate_synthetics_event([], [], should_exist=False)
@override_application_settings(_disabled_settings)
def test_synthetics_disabled():
    headers = _make_synthetics_headers()
    response = target_application.get("/", headers=headers)


_external_synthetics_headers = _make_synthetics_headers()
_external_synthetics_header = _external_synthetics_headers["X-NewRelic-Synthetics"]
_external_synthetics_info_header = _external_synthetics_headers["X-NewRelic-Synthetics-Info"]


@cat_enabled
@validate_synthetics_external_trace_header(_external_synthetics_header, _external_synthetics_info_header)
@override_application_settings(_override_settings)
def test_valid_synthetics_external_trace_header():
    headers = _make_synthetics_headers()
    response = target_application.get("/", headers=headers)


@cat_enabled
@validate_synthetics_external_trace_header(_external_synthetics_header, None)
@override_application_settings(_override_settings)
def test_valid_synthetics_external_trace_header_without_info():
    headers = _make_synthetics_headers(type_=None)
    response = target_application.get("/", headers=headers)


@cat_enabled
@validate_synthetics_external_trace_header(_external_synthetics_header, _external_synthetics_info_header)
@override_application_settings(_override_settings)
def test_valid_external_trace_header_with_byte_inbound_header():
    headers = _make_synthetics_headers()
    headers = {k.encode("utf-8"): v.encode("utf-8") for k, v in headers.items()}

    @web_transaction(
        name="test_valid_external_trace_header_with_byte_inbound_header",
        headers=headers,
    )
    def webapp():
        pass

    webapp()


@validate_synthetics_external_trace_header(None, None)
@override_application_settings(_override_settings)
def test_no_synthetics_external_trace_header():
    response = target_application.get("/")


def _synthetics_limit_test(num_requests, num_events, num_transactions):

    # Force harvest to clear stats

    instance = agent_instance()
    application = list(instance.applications.values())[0]
    application.harvest()

    # Send requests

    headers = _make_synthetics_headers()
    for i in range(num_requests):
        response = target_application.get("/", headers=headers)

    # Check that we've saved the right number events and traces

    stats = application._stats_engine
    assert len(stats.synthetics_events) == num_events
    assert len(stats.synthetics_transactions) == num_transactions


@pytest.mark.skipif(True, reason="Test is too flaky. Need to find a way to make harvests more predictable.")
@pytest.mark.parametrize(
    "num_requests,num_events,num_transactions", [(0, 0, 0), (20, 20, 20), (21, 21, 20), (200, 200, 20), (201, 200, 20)]
)
@override_application_settings(_override_settings)
def test_synthetics_requests_default_limits(num_requests, num_events, num_transactions):
    _synthetics_limit_test(num_requests, num_events, num_transactions)


_custom_settings = {
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [int(ACCOUNT_ID)],
    "agent_limits.synthetics_events": 5,
    "agent_limits.synthetics_transactions": 3,
    "synthetics.enabled": True,
}


@pytest.mark.skipif(True, reason="Test is too flaky. Need to find a way to make harvests more predictable.")
@pytest.mark.parametrize(
    "num_requests,num_events,num_transactions", [(0, 0, 0), (3, 3, 3), (4, 4, 3), (5, 5, 3), (6, 5, 3)]
)
@override_application_settings(_custom_settings)
def test_synthetics_requests_custom_limits(num_requests, num_events, num_transactions):
    _synthetics_limit_test(num_requests, num_events, num_transactions)


_zero_settings = {
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [int(ACCOUNT_ID)],
    "agent_limits.synthetics_events": 0,
    "agent_limits.synthetics_transactions": 0,
    "synthetics.enabled": True,
}


@pytest.mark.skipif(True, reason="Test is too flaky. Need to find a way to make harvests more predictable.")
@pytest.mark.parametrize("num_requests,num_events,num_transactions", [(0, 0, 0), (1, 0, 0)])
@override_application_settings(_zero_settings)
def test_synthetics_requests_zero_limits(num_requests, num_events, num_transactions):
    _synthetics_limit_test(num_requests, num_events, num_transactions)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import catch_background_exceptions
from newrelic.common.object_wrapper import (
        transient_function_wrapper,
        function_wrapper)


def validate_apdex_metrics(name, group='Function', count=1,
        apdex_t_min=0.5, apdex_t_max=0.5):

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        record_transaction_called = []
        recorded_metrics = []

        @transient_function_wrapper('newrelic.core.stats_engine',
                'StatsEngine.record_transaction')
        @catch_background_exceptions
        def _capture_metrics(wrapped, instance, args, kwargs):
            record_transaction_called.append(True)
            result = wrapped(*args, **kwargs)
            recorded_metrics.append(instance.stats_table)
            return result

        def _validate():
            metric_name = 'Apdex/%s/%s' % (group, name)
            key = (metric_name, '')

            metric = recorded_metrics.get(key)

            # Sum of satisfying / tolerating / frustrating must equal count
            assert sum(metric[:2]) == count, metric[:2]

            # Check apdex_t range
            assert metric[3] == apdex_t_min, metric[3]
            assert metric[4] == apdex_t_max, metric[4]

            # Last field is always 0
            assert metric[5] == 0

        _new_wrapper = _capture_metrics(wrapped)
        result = _new_wrapper(*args, **kwargs)
        assert record_transaction_called
        recorded_metrics = recorded_metrics.pop()
        _validate()
        return result

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

from newrelic.api.transaction import current_transaction


def native_coroutine_test(
    event_loop, transaction, nr_enabled=True, does_hang=False, call_exit=False, runtime_error=False
):
    @transaction
    async def task():
        txn = current_transaction()

        if not nr_enabled:
            assert txn is None

        if call_exit:
            txn.__exit__(None, None, None)
        else:
            assert current_transaction() is txn

        try:
            if does_hang:
                await event_loop.create_future()
            else:
                await asyncio.sleep(0.0)
        except GeneratorExit:
            if runtime_error:
                await asyncio.sleep(0.0)

    return task



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture()
def existing_document(collection):
    doc = collection.document("document")
    doc.set({"x": 1})
    return doc


@pytest.fixture()
def exercise_async_client(async_client, existing_document):
    async def _exercise_async_client():
        assert len([_ async for _ in async_client.collections()]) >= 1
        doc = [_ async for _ in async_client.get_all([existing_document])][0]
        assert doc.to_dict()["x"] == 1

    return _exercise_async_client


def test_firestore_async_client(loop, exercise_async_client, instance_info):
    _test_scoped_metrics = [
        ("Datastore/operation/Firestore/collections", 1),
        ("Datastore/operation/Firestore/get_all", 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_client",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_client")
    def _test():
        loop.run_until_complete(exercise_async_client())

    _test()


@background_task()
def test_firestore_async_client_generators(async_client, collection, assert_trace_for_async_generator):
    doc = collection.document("test")
    doc.set({})

    assert_trace_for_async_generator(async_client.collections)
    assert_trace_for_async_generator(async_client.get_all, [doc])


def test_firestore_async_client_trace_node_datastore_params(loop, exercise_async_client, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_client())

    _test()


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "application_logging.enabled": True,
    "application_logging.forwarding.enabled": True,
    "application_logging.metrics.enabled": True,
    "application_logging.local_decorating.enabled": True,
    "event_harvest_config.harvest_limits.log_event_data": 100000,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (logger_loguru)",
    default_settings=_default_settings,
)


class CaplogHandler(logging.StreamHandler):
    """
    To prevent possible issues with pytest's monkey patching
    use a custom Caplog handler to capture all records
    """
    def __init__(self, *args, **kwargs):
        self.records = []
        super(CaplogHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        self.records.append(self.format(record))


@pytest.fixture(scope="function")
def logger():
    import loguru
    _logger = loguru.logger
    caplog = CaplogHandler()
    handler_id = _logger.add(caplog, level="WARNING", format="{message}")
    _logger.caplog = caplog
    yield _logger
    del caplog.records[:]
    _logger.remove(handler_id)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.core.exceptions import ObjectDoesNotExist
from tastypie.resources import Resource
from tastypie.exceptions import NotFound


class SimpleResource(Resource):
    class Meta:
        resource_name = 'simple'

    def obj_get(self, *args, **kwargs):
        pk = kwargs['pk']
        if pk == 'NotFound':
            raise NotFound('Object not found.')
        elif pk == 'ObjectDoesNotExist':
            raise ObjectDoesNotExist('Object does not exist.')
        elif pk == 'ZeroDivisionError':
            1 / 0
        else:
            raise NotImplemented()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def test_pywsgi_application_index(target_application):
    for i in range(3):
        response = target_application.get('/pywsgi')
        response.mustcontain('WSGI RESPONSE')

def test_pywsgi_request_timeout_application(target_application):
    for i in range(3):
        response = target_application.get(
                '/request-timeout-application/pywsgi', status=500)

def test_pywsgi_request_timeout_response(target_application):
    # The gevent pywsgi server appears to not be WSGI compliant in this
    # test. The expectation in yielding a string and then causing a
    # timeout would be that the HTTP 200 OK response and initial text
    # would have already been sent. As such we should always see a HTTP
    # 200 OK response and partial content. For some reason when the test
    # client is in the same process as the server this doesn't happen.
    # If one uses an extenal browser against the pywsgi server it works
    # as expected. It may all be done to some issue with how coroutines
    # are schedule when done within the one process.

    for i in range(3):
        response = target_application.get(
                '/request-timeout-response/pywsgi', status=500)
        #response.mustcontain('WSGI')

def test_pywsgi_request_timeout_finalize(target_application):
    # This suffers same issue as in test_pywsgi_request_timeout_response()
    # where the results isn't what we expect when things run in the same
    # process.

    for i in range(3):
        response = target_application.get(
                '/request-timeout-finalize/pywsgi', status=500)
        #response.mustcontain('WSGI RESPONSE')

def test_pywsgi_raise_exception_application(target_application):
    for i in range(3):
        response = target_application.get(
                '/raise-exception-application/pywsgi', status=500)

def test_pywsgi_raise_exception_response(target_application):
    for i in range(3):
        response = target_application.get(
                '/raise-exception-response/pywsgi', status=500)

def test_pywsgi_raise_exception_finalize(target_application):
    for i in range(3):
        response = target_application.get(
                '/raise-exception-finalize/pywsgi', status=500)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import wrap_datastore_trace
from newrelic.api.function_trace import wrap_function_trace
from newrelic.common.async_wrapper import generator_wrapper, async_generator_wrapper


def _conn_str_to_host(getter):
    """Safely transform a getter that can retrieve a connection string into the resulting host."""

    def closure(obj, *args, **kwargs):
        try:
            return getter(obj, *args, **kwargs).split(":")[0]
        except Exception:
            return None

    return closure


def _conn_str_to_port(getter):
    """Safely transform a getter that can retrieve a connection string into the resulting port."""

    def closure(obj, *args, **kwargs):
        try:
            return getter(obj, *args, **kwargs).split(":")[1]
        except Exception:
            return None

    return closure


# Default Target ID and Instance Info
_get_object_id = lambda obj, *args, **kwargs: getattr(obj, "id", None)
_get_client_database_string = lambda obj, *args, **kwargs: getattr(
    getattr(obj, "_client", None), "_database_string", None
)
_get_client_target = lambda obj, *args, **kwargs: obj._client._target
_get_client_target_host = _conn_str_to_host(_get_client_target)
_get_client_target_port = _conn_str_to_port(_get_client_target)

# Client Instance Info
_get_database_string = lambda obj, *args, **kwargs: getattr(obj, "_database_string", None)
_get_target = lambda obj, *args, **kwargs: obj._target
_get_target_host = _conn_str_to_host(_get_target)
_get_target_port = _conn_str_to_port(_get_target)

# Query Target ID
_get_parent_id = lambda obj, *args, **kwargs: getattr(getattr(obj, "_parent", None), "id", None)

# AggregationQuery Target ID
_get_collection_ref_id = lambda obj, *args, **kwargs: getattr(getattr(obj, "_collection_ref", None), "id", None)


def instrument_google_cloud_firestore_v1_base_client(module):
    rollup = ("Datastore/all", "Datastore/Firestore/all")
    wrap_function_trace(
        module, "BaseClient.__init__", name="%s:BaseClient.__init__" % module.__name__, terminal=True, rollup=rollup
    )


def instrument_google_cloud_firestore_v1_client(module):
    if hasattr(module, "Client"):
        class_ = module.Client
        for method in ("collections", "get_all"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "Client.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=None,
                    host=_get_target_host,
                    port_path_or_id=_get_target_port,
                    database_name=_get_database_string,
                    async_wrapper=generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_async_client(module):
    if hasattr(module, "AsyncClient"):
        class_ = module.AsyncClient
        for method in ("collections", "get_all"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncClient.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=None,
                    host=_get_target_host,
                    port_path_or_id=_get_target_port,
                    database_name=_get_database_string,
                    async_wrapper=async_generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_collection(module):
    if hasattr(module, "CollectionReference"):
        class_ = module.CollectionReference
        for method in ("add", "get"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "CollectionReference.%s" % method,
                    product="Firestore",
                    target=_get_object_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("stream", "list_documents"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "CollectionReference.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_object_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_async_collection(module):
    if hasattr(module, "AsyncCollectionReference"):
        class_ = module.AsyncCollectionReference
        for method in ("add", "get"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncCollectionReference.%s" % method,
                    product="Firestore",
                    target=_get_object_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    operation=method,
                )

        for method in ("stream", "list_documents"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncCollectionReference.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_object_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=async_generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_document(module):
    if hasattr(module, "DocumentReference"):
        class_ = module.DocumentReference
        for method in ("create", "delete", "get", "set", "update"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "DocumentReference.%s" % method,
                    product="Firestore",
                    target=_get_object_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("collections",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "DocumentReference.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_object_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_async_document(module):
    if hasattr(module, "AsyncDocumentReference"):
        class_ = module.AsyncDocumentReference
        for method in ("create", "delete", "get", "set", "update"):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncDocumentReference.%s" % method,
                    product="Firestore",
                    target=_get_object_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("collections",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncDocumentReference.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_object_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=async_generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_query(module):
    if hasattr(module, "Query"):
        class_ = module.Query
        for method in ("get",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "Query.%s" % method,
                    product="Firestore",
                    target=_get_parent_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("stream",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "Query.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_parent_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=generator_wrapper,
                )

    if hasattr(module, "CollectionGroup"):
        class_ = module.CollectionGroup
        for method in ("get_partitions",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "CollectionGroup.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_parent_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_async_query(module):
    if hasattr(module, "AsyncQuery"):
        class_ = module.AsyncQuery
        for method in ("get",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncQuery.%s" % method,
                    product="Firestore",
                    target=_get_parent_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("stream",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncQuery.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_parent_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=async_generator_wrapper,
                )

    if hasattr(module, "AsyncCollectionGroup"):
        class_ = module.AsyncCollectionGroup
        for method in ("get_partitions",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncCollectionGroup.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_parent_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=async_generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_aggregation(module):
    if hasattr(module, "AggregationQuery"):
        class_ = module.AggregationQuery
        for method in ("get",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AggregationQuery.%s" % method,
                    product="Firestore",
                    target=_get_collection_ref_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("stream",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AggregationQuery.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_collection_ref_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_async_aggregation(module):
    if hasattr(module, "AsyncAggregationQuery"):
        class_ = module.AsyncAggregationQuery
        for method in ("get",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncAggregationQuery.%s" % method,
                    product="Firestore",
                    target=_get_collection_ref_id,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )

        for method in ("stream",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncAggregationQuery.%s" % method,
                    operation=method,
                    product="Firestore",
                    target=_get_collection_ref_id,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                    async_wrapper=async_generator_wrapper,
                )


def instrument_google_cloud_firestore_v1_batch(module):
    if hasattr(module, "WriteBatch"):
        class_ = module.WriteBatch
        for method in ("commit",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "WriteBatch.%s" % method,
                    product="Firestore",
                    target=None,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )


def instrument_google_cloud_firestore_v1_async_batch(module):
    if hasattr(module, "AsyncWriteBatch"):
        class_ = module.AsyncWriteBatch
        for method in ("commit",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "AsyncWriteBatch.%s" % method,
                    product="Firestore",
                    target=None,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )


def instrument_google_cloud_firestore_v1_bulk_batch(module):
    if hasattr(module, "BulkWriteBatch"):
        class_ = module.BulkWriteBatch
        for method in ("commit",):
            if hasattr(class_, method):
                wrap_datastore_trace(
                    module,
                    "BulkWriteBatch.%s" % method,
                    product="Firestore",
                    target=None,
                    operation=method,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )


def instrument_google_cloud_firestore_v1_transaction(module):
    if hasattr(module, "Transaction"):
        class_ = module.Transaction
        for method in ("_commit", "_rollback"):
            if hasattr(class_, method):
                operation = method[1:]  # Trim leading underscore
                wrap_datastore_trace(
                    module,
                    "Transaction.%s" % method,
                    product="Firestore",
                    target=None,
                    operation=operation,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )


def instrument_google_cloud_firestore_v1_async_transaction(module):
    if hasattr(module, "AsyncTransaction"):
        class_ = module.AsyncTransaction
        for method in ("_commit", "_rollback"):
            if hasattr(class_, method):
                operation = method[1:]  # Trim leading underscore
                wrap_datastore_trace(
                    module,
                    "AsyncTransaction.%s" % method,
                    product="Firestore",
                    target=None,
                    operation=operation,
                    host=_get_client_target_host,
                    port_path_or_id=_get_client_target_port,
                    database_name=_get_client_database_string,
                )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import botocore.session
import moto
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

MOTO_VERSION = get_package_version_tuple("moto")

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re

    moto.packages.responses.responses.re._pattern_type = re.Pattern

AWS_ACCESS_KEY_ID = "AAAAAAAAAAAACCESSKEY"
AWS_SECRET_ACCESS_KEY = "AAAAAASECRETKEY"  # nosec (This is fine for testing purposes)
AWS_REGION = "us-east-1"
UBUNTU_14_04_PARAVIRTUAL_AMI = "ami-c65be9ae"

TEST_INSTANCE = "python-agent-test-%s" % uuid.uuid4()

_ec2_scoped_metrics = [
    ("External/ec2.us-east-1.amazonaws.com/botocore/POST", 3),
]

_ec2_rollup_metrics = [
    ("External/all", 3),
    ("External/allOther", 3),
    ("External/ec2.us-east-1.amazonaws.com/all", 3),
    ("External/ec2.us-east-1.amazonaws.com/botocore/POST", 3),
]


@dt_enabled
@validate_span_events(expected_agents=("aws.requestId",), count=3)
@validate_span_events(exact_agents={"aws.operation": "RunInstances"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DescribeInstances"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "TerminateInstances"}, count=1)
@validate_tt_segment_params(present_params=("aws.requestId",))
@validate_transaction_metrics(
    "test_botocore_ec2:test_ec2",
    scoped_metrics=_ec2_scoped_metrics,
    rollup_metrics=_ec2_rollup_metrics,
    background_task=True,
)
@background_task()
@moto.mock_ec2
def test_ec2():
    session = botocore.session.get_session()
    client = session.create_client(
        "ec2", region_name=AWS_REGION, aws_access_key_id=AWS_ACCESS_KEY_ID, aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )

    # Create instance
    resp = client.run_instances(
        ImageId=UBUNTU_14_04_PARAVIRTUAL_AMI,
        InstanceType="m1.small",
        MinCount=1,
        MaxCount=1,
    )
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert len(resp["Instances"]) == 1
    instance_id = resp["Instances"][0]["InstanceId"]

    # Describe instance
    resp = client.describe_instances(InstanceIds=[instance_id])
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert resp["Reservations"][0]["Instances"][0]["InstanceId"] == instance_id

    # Delete instance
    resp = client.terminate_instances(InstanceIds=[instance_id])
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert resp["TerminatingInstances"][0]["InstanceId"] == instance_id
